@@ -41,17 +41,59 @@
 
 #include <linux/elf.h>
 
+#ifdef ELF_TRACE
+#include <ibcs/trace.h>
+#endif
+
+
+#undef ELF_PARSE_COMMENTS
+
 
 static int load_elf_binary(struct linux_binprm * bprm, struct pt_regs * regs);
 static int load_elf_library(int fd);
 
 
+#ifdef ELF_PARSE_COMMENTS
+static int parse_comments(struct linux_binprm *exe_bprm, Elf32_Shdr * sect);
+
+/*
+ *  The following table gives clues to the "personality" of the executable
+ *  which we hope to find in the .comment sections of the binary.
+ *  The set here may not be comprehensive even for those systems listed.
+ *  Use 'mcs -p' to list the .comments sections of a binary to see what
+ *  clues might be there. Or use 'strings' if you don't have mcs.
+ */
+struct elf_clue {
+    short len;		/* negative number uses strstr for lookup */
+    char *text;
+    unsigned long personality;
+};
+
+static struct elf_clue elf_clues[] = {
+    /* SCO Unix OpenServer 5 */
+    { 18, "@(#)SCO OpenServer",				PER_SCOSVR3 },
+    { 22, "@(#) crt1.s.source 20.",			PER_SCOSVR3 },
+#if 0
+    /* Interactive SVR4 */
+    { 33, "@(#)UNIX System V/386 Release 4.0",		PER_SVR4 },
+#endif
+#if 0
+    /* UnixWare 1.x (at _end_ of string) */
+    {  -7, " Univel",					PER_SVR4 },
+#endif
+
+    /* End of table */
+    { 0, 0, 0 }
+};
+#endif /* ELF_PARSE_COMMENTS */
+
+
 #ifdef INIT_MM
-struct linux_binfmt elf_format = {
+struct linux_binfmt ibcs_elf_format = {
 	NULL, &mod_use_count_, load_elf_binary, load_elf_library, NULL
 };
 #else
-struct linux_binfmt elf_format = {
+struct linux_binfmt ibcs_elf_format = {
 	load_elf_binary, load_elf_library
 };
 #endif
@@ -62,7 +104,9 @@ struct linux_binfmt elf_format = {
    contain the junk from the file that should not
    be in memory */
 
-static void padzero(unsigned long elf_bss){
+static void
+padzero(unsigned long elf_bss)
+{
 	unsigned int nbyte;
 	char *fpnt;
   
@@ -81,11 +125,16 @@ unsigned long * create_elf_tables(char * p,int argc,int envc,struct elfhdr * exe
 {
 	unsigned long *argv,*envp, *dlinfo;
 	unsigned long * sp;
+#if LINUX_VERSION_CODE < 66375
 	struct vm_area_struct *mpnt;
 
 	mpnt = (struct vm_area_struct *)kmalloc(sizeof(*mpnt), GFP_KERNEL);
 	if (mpnt) {
+#ifdef __NR_getsid
+		mpnt->vm_mm = current->mm;
+#else
 		mpnt->vm_task = current;
+#endif
 		mpnt->vm_start = PAGE_MASK & (unsigned long) p;
 		mpnt->vm_end = TASK_SIZE;
 		mpnt->vm_page_prot = PAGE_COPY;
@@ -104,10 +153,14 @@ unsigned long * create_elf_tables(char * p,int argc,int envc,struct elfhdr * exe
 		mpnt->vm_offset = 0;
 		mpnt->vm_ops = NULL;
 		insert_vm_struct(current, mpnt);
+#ifdef __NR_mlock
+		current->MM(total_vm) += (mpnt->vm_end - mpnt->vm_start) >> PAGE_SHIFT;
+#endif
 #ifndef VM_GROWSDOWN
 		current->MM(stk_vma) = mpnt;
 #endif
 	}
+#endif
 	sp = (unsigned long *) (0xfffffffc & (unsigned long) p);
 	if(exec) sp -= DLINFO_ITEMS*2;
 	dlinfo = sp;
@@ -171,11 +224,12 @@ static unsigned int load_elf_interp(struct elfhdr * interp_elf_ex,
 	int elf_bss;
 #ifndef STACK_TOP
 	int old_fs;
-#else
+#endif
 	int retval;
 	unsigned int last_bss;
 	int error;
-	int i, k;
+	int i;
+	unsigned int k;
 	
 	elf_bss = 0;
 	last_bss = 0;
@@ -218,32 +272,34 @@ static unsigned int load_elf_interp(struct elfhdr * interp_elf_ex,
 	for(i=0; i<interp_elf_ex->e_phnum; i++, eppnt++)
 	  if(eppnt->p_type == PT_LOAD) {
 #ifdef PF_R
-		int elf_type = MAP_PRIVATE | MAP_DENY_WRITE;
+		int elf_type = MAP_PRIVATE | MAP_DENYWRITE;
 		int elf_prot = 0;
 		unsigned long vaddr = 0;
 	    if (eppnt->p_flags & PF_R) elf_prot = PROT_READ;
 	    if (eppnt->p_flags & PF_W) elf_prot |= PROT_WRITE;
 	    if (eppnt->p_flags & PF_X) elf_prot |= PROT_EXEC;
-		if (interp_elf_ex->e_type == ET_EXEC) {
+		if (interp_elf_ex->e_type == ET_EXEC || load_addr != 0) {
 			elf_type |= MAP_FIXED;
 			vaddr = eppnt->p_vaddr;
 	    }
 	    error = do_mmap(file, 
-			    vaddr & 0xfffff000,
+			    load_addr + (vaddr & 0xfffff000),
 			    eppnt->p_filesz + (vaddr & 0xfff),
 				elf_prot,
 				elf_type,
 			    eppnt->p_offset & 0xfffff000);
 #else
 	    error = do_mmap(file, 
-			    eppnt->p_vaddr & 0xfffff000,
+			    load_addr + (eppnt->p_vaddr & 0xfffff000),
 			    eppnt->p_filesz + (eppnt->p_vaddr & 0xfff),
 			    PROT_READ | PROT_WRITE | PROT_EXEC,
 #ifdef MAP_EXECUTABLE
 			    MAP_PRIVATE | MAP_DENYWRITE | MAP_EXECUTABLE | (interp_elf_ex->e_type == ET_EXEC ? MAP_FIXED : 0),
 #else
 #  ifdef MAP_DENYWRITE
-			    MAP_PRIVATE | MAP_DENYWRITE | (interp_elf_ex->e_type == ET_EXEC ? MAP_FIXED : 0),
+			    MAP_PRIVATE | MAP_DENYWRITE |
+			    ((interp_elf_ex->e_type == ET_EXEC || load_addr)
+				? MAP_FIXED : 0),
 #  else
 			    MAP_PRIVATE | (interp_elf_ex->e_type == ET_EXEC ? MAP_FIXED : 0),
 #  endif
@@ -251,26 +307,16 @@ static unsigned int load_elf_interp(struct elfhdr * interp_elf_ex,
 			    eppnt->p_offset & 0xfffff000);
 #endif /* PF_R */
 	    
-		if (error < 0 && error > -1024) break;  /* Real error */
-
 	    if(!load_addr && interp_elf_ex->e_type == ET_DYN)
 	      load_addr = error;
-
-		/*
-		 * Find the end of file mapping for this phdr, and keep
-		 * track of the largest address we see for this.
-		 */
 	    k = load_addr + eppnt->p_vaddr + eppnt->p_filesz;
 	    if(k > elf_bss) elf_bss = k;
-
-		/*
-		 * Do the same thing for the memory mapping - between
-		 * elf_bss and last_bss is the bss section.
-		 */
+	    if (error < 0 && error > -1024) break;  /* Real error */
 	    k = load_addr + eppnt->p_memsz + eppnt->p_vaddr;
 	    if(k > last_bss) last_bss = k;
 	  }
 	
+
 	/* Now use mmap to map the library into memory. */
 
 	SYS(close)(elf_exec_fileno);
@@ -279,12 +325,6 @@ static unsigned int load_elf_interp(struct elfhdr * interp_elf_ex,
 		return 0xffffffff;
 	}
 
-	/*
-	 * Now fill out the bss section. First pad the last page up
-	 * to the page boundary, and then perform a mmap to make sure
-	 * that there are zeromapped pages up to and including the
-	 * last bss page.
-	 */
 	padzero(elf_bss);
 	len = (elf_bss + 0xfff) & 0xfffff000; /* What we have mapped so far */
 
@@ -304,7 +344,7 @@ static unsigned int load_aout_interp(struct exec * interp_ex,
 {
   int retval;
   unsigned int elf_entry;
-  
+
   current->MM(brk) = interp_ex->a_bss +
     (current->MM(end_data) = interp_ex->a_data +
      (current->MM(end_code) = interp_ex->a_text));
@@ -330,7 +370,7 @@ static unsigned int load_aout_interp(struct exec * interp_ex,
     retval = read_exec(interpreter_inode,
 		       N_TXTOFF(*interp_ex) ,
 		       (char *) N_TXTADDR(*interp_ex),
-		       interp_ex->a_text+interp_ex->a_data), 0;
+		       interp_ex->a_text+interp_ex->a_data, 0);
 #else
     retval = read_exec(interpreter_inode,
 		       N_TXTOFF(*interp_ex) ,
@@ -405,6 +445,68 @@ static int load_elf_binary(struct linux_binprm * bprm, struct pt_regs * regs)
 		return -ENOEXEC;
 	}
 	
+	current->personality = PER_LINUX;
+
+#ifdef ELF_PARSE_COMMENTS
+	/* First scan any sections for .comments which may have
+	 * personality clues.
+	 */
+	if (elf_ex.e_shoff) {
+		Elf32_Shdr * elf_spnt, *elf_shdata;
+
+		elf_shdata = (Elf32_Shdr *)kmalloc(elf_ex.e_shentsize *
+						   elf_ex.e_shnum, GFP_KERNEL);
+#ifdef STACK_TOP
+		retval = read_exec(bprm->inode, elf_ex.e_shoff,
+				(char *) elf_shdata,
+		 		elf_ex.e_shentsize * elf_ex.e_shnum, 1);
+#else
+		old_fs = get_fs();
+		set_fs(get_ds());
+		retval = read_exec(bprm->inode, elf_ex.e_shoff,
+				(char *) elf_shdata,
+				elf_ex.e_shentsize * elf_ex.e_shnum);
+		set_fs(old_fs);
+#endif
+		if (retval < 0) {
+	        	kfree(elf_shdata);
+			MOD_DEC_USE_COUNT;
+			return retval;
+		}
+
+		elf_spnt = elf_shdata;
+		for(i=0; retval && i < elf_ex.e_shnum; i++){
+			if (elf_spnt->sh_type == SHT_PROGBITS
+			&& elf_spnt->sh_flags == 0
+			&& (status = parse_comments(bprm, elf_spnt)) > 0)
+				break;
+			elf_spnt++;
+			retval -= elf_ex.e_shentsize;
+		}
+
+		kfree(elf_shdata);
+	}
+#endif /* ELF_PARSE_COMMENTS */
+
+#ifdef CONFIG_BINFMT_ELF
+	/* If we didn't recognise it as being other than Linux return
+	 * ENOEXEC and let the standard Linux ELF loader handle it.
+	 * Otherwise we will save a reference to this loader for each
+	 * Linux ELF binary which either risks a kernel crash if we
+	 * unload iBCS or prevents us unloading iBCS entirely.
+	 */
+	if (current->personality == PER_LINUX) {
+#ifdef ELF_TRACE
+		if ((ibcs_trace & TRACE_ELF_LD)) {
+			printk(KERN_DEBUG "ELF: Linux or generic SVR4"
+					" - hand off to kernel loader.\n");
+		}
+#endif
+		MOD_DEC_USE_COUNT;
+		return -ENOEXEC;
+	}
+#endif
+
 	/* Now read in all of the header information */
 	
 	elf_phdata = (struct elf_phdr *) kmalloc(elf_ex.e_phentsize * 
@@ -443,7 +545,7 @@ static int load_elf_binary(struct linux_binprm * bprm, struct pt_regs * regs)
 	
 	elf_stack = 0xffffffff;
 	elf_interpreter = NULL;
-	start_code = 0;
+	start_code = 0xffffffff;
 	end_code = 0;
 	end_data = 0;
 	
@@ -459,6 +561,16 @@ static int load_elf_binary(struct linux_binprm * bprm, struct pt_regs * regs)
 			
 			elf_interpreter = (char *) kmalloc(elf_ppnt->p_filesz, 
 							   GFP_KERNEL);
+			if (elf_interpreter == NULL) {
+#ifndef STACK_TOP
+				set_fs(old_fs);
+#endif
+				kfree (elf_phdata);
+				kfree(elf_interpreter);
+				SYS(close)(elf_exec_fileno);
+				MOD_DEC_USE_COUNT;
+				return -ENOMEM;
+			}
 			
 #ifdef STACK_TOP
 			retval = read_exec(bprm->inode,elf_ppnt->p_offset,elf_interpreter,
@@ -499,8 +611,12 @@ static int load_elf_binary(struct linux_binprm * bprm, struct pt_regs * regs)
 				
 			}
 			if(retval < 0) {
+#ifndef STACK_TOP
+				set_fs(old_fs);
+#endif
 				kfree (elf_phdata);
 				kfree(elf_interpreter);
+				SYS(close)(elf_exec_fileno);
 				MOD_DEC_USE_COUNT;
 				return retval;
 			}
@@ -518,6 +634,7 @@ static int load_elf_binary(struct linux_binprm * bprm, struct pt_regs * regs)
 		if(retval < 0) {
 			kfree(elf_interpreter);
 			kfree(elf_phdata);
+			SYS(close)(elf_exec_fileno);
 			MOD_DEC_USE_COUNT;
 			return -ELIBACC;
 		}
@@ -535,6 +652,7 @@ static int load_elf_binary(struct linux_binprm * bprm, struct pt_regs * regs)
 		  {
 		    kfree(elf_interpreter);
 		    kfree(elf_phdata);
+		    SYS(close)(elf_exec_fileno);
 		    MOD_DEC_USE_COUNT;
 		    return -ELIBBAD;
 		  }
@@ -560,6 +678,7 @@ static int load_elf_binary(struct linux_binprm * bprm, struct pt_regs * regs)
 			      kfree(elf_interpreter);
 			}
 		        kfree (elf_phdata);
+			SYS(close)(elf_exec_fileno);
 			MOD_DEC_USE_COUNT;
 			return -E2BIG;
 		}
@@ -577,8 +696,12 @@ static int load_elf_binary(struct linux_binprm * bprm, struct pt_regs * regs)
 	/* Do this so that we can load the interpreter, if need be.  We will
 	   change some of these later */
 	current->MM(rss) = 0;
-#ifdef STACK_TOP
+#if LINUX_VERSION_CODE < 66375
+#ifndef STACK_TOP
 	bprm->p += setup_arg_pages(0, bprm->page);
+#endif
+#else
+	bprm->p = setup_arg_pages(bprm->p, bprm);
 #endif
 	current->MM(start_stack) = bprm->p;
 	
@@ -656,10 +779,14 @@ static int load_elf_binary(struct linux_binprm * bprm, struct pt_regs * regs)
 			if(!load_addr) 
 			  load_addr = elf_ppnt->p_vaddr - elf_ppnt->p_offset;
 			k = elf_ppnt->p_vaddr;
-			if(k > start_code) start_code = k;
+			if(k < start_code) start_code = k;
 			k = elf_ppnt->p_vaddr + elf_ppnt->p_filesz;
 			if(k > elf_bss) elf_bss = k;
+#ifdef PF_X
+			if((elf_ppnt->p_flags & PF_X) && end_code <  k)
+#else
 			if((elf_ppnt->p_flags | PROT_WRITE) && end_code <  k)
+#endif
 				end_code = k; 
 			if(end_data < k) end_data = k; 
 			k = elf_ppnt->p_vaddr + elf_ppnt->p_memsz;
@@ -672,14 +799,15 @@ static int load_elf_binary(struct linux_binprm * bprm, struct pt_regs * regs)
 	kfree(elf_phdata);
 	
 	if(interpreter_type != INTERPRETER_AOUT) SYS(close)(elf_exec_fileno);
-	current->personality = (ibcs2_interpreter ? PER_SVR4 : PER_LINUX);
+	if (current->personality == PER_LINUX && ibcs2_interpreter)
+		current->personality = PER_SVR4;
 #ifdef INIT_MM
 	if (current->exec_domain && current->exec_domain->use_count)
 		(*current->exec_domain->use_count)--;
 	if (current->binfmt && current->binfmt->use_count)
 		(*current->binfmt->use_count)--;
 	current->exec_domain = lookup_exec_domain(current->personality);
-	current->binfmt = &elf_format;
+	current->binfmt = &ibcs_elf_format;
 	if (current->exec_domain && current->exec_domain->use_count)
 		(*current->exec_domain->use_count)++;
 	if (current->binfmt && current->binfmt->use_count)
@@ -690,9 +818,11 @@ static int load_elf_binary(struct linux_binprm * bprm, struct pt_regs * regs)
 	bprm->inode->i_count++;
 #endif
 #ifdef LOW_ELF_STACK
-	current->start_stack = p = elf_stack - 4;
+	current->start_stack = bprm->p = elf_stack - 4;
 #endif
+#if LINUX_VERSION_CODE < 66375
 	bprm->p -= MAX_ARG_PAGES*PAGE_SIZE;
+#endif
 	bprm->p = (unsigned long) 
 	  create_elf_tables((char *)bprm->p,
 			bprm->argc,
@@ -715,6 +845,9 @@ static int load_elf_binary(struct linux_binprm * bprm, struct pt_regs * regs)
 	current->suid = current->euid = bprm->e_uid;
 	current->sgid = current->egid = bprm->e_gid;
 #endif
+#ifdef PF_FORKNOEXEC
+	current->flags &= ~PF_FORKNOEXEC;
+#endif
 
 	/* Calling sys_brk effectively mmaps the pages that we need for the bss and break
 	   sections */
@@ -723,12 +856,23 @@ static int load_elf_binary(struct linux_binprm * bprm, struct pt_regs * regs)
 
 	padzero(elf_bss);
 
-	/* Why this, you ask???  Well SVr4 maps page 0 as read-only,
-	   and some applications "depend" upon this behavior.
-	   Since we do not have the power to recompile these, we
-	   emulate the SVr4 behavior.  Sigh.  */
-	error = do_mmap(NULL, 0, 4096, PROT_READ | PROT_EXEC,
-			MAP_FIXED | MAP_PRIVATE, 0);
+	if (current->personality != PER_LINUX) {
+		/* Why this, you ask???  Well SVr4 maps page 0 as read-only,
+		   and some applications "depend" upon this behavior.
+		   Since we do not have the power to recompile these, we
+		   emulate the SVr4 behavior.  Sigh.  */
+		error = do_mmap(NULL, 0, 4096, PROT_READ | PROT_EXEC,
+				MAP_FIXED | MAP_PRIVATE, 0);
+	}
+
+	/* SVR4/i386 ABI (pages 3-31, 3-32) says that when the program
+	   starts %edx contains a pointer to a function which might be
+	   registered using `atexit'.  This provides a mean for the
+	   dynamic linker to call DT_FINI functions for shared libraries
+	   that have been loaded before the code runs.
+
+	   A value of 0 tells we have no such handler.  */
+	regs->edx = 0;
 
 	start_thread(regs, elf_entry, bprm->p);
 	if (current->flags & PF_PTRACED)
@@ -737,6 +881,115 @@ static int load_elf_binary(struct linux_binprm * bprm, struct pt_regs * regs)
 	return 0;
 }
 
+
+#ifdef ELF_PARSE_COMMENTS
+/*
+ *  Parse a comments section looking for clues as to the system this
+ *  was compiled on so we can get the system call interface right.
+ */
+static int
+parse_comments(struct linux_binprm *exe_bprm, Elf32_Shdr * sect)
+{
+	unsigned long offset, nbytes;
+	char *buffer;
+	int old_fs;
+	int status;
+
+	/* Fetch the size of the section. There must be something in there
+	 * or the section wouldn't exist at all. We only bother with the
+	 * first 8192 characters though. There isn't any point getting too
+	 * carried away!
+	 */
+	if ((nbytes = sect->sh_size) > 8192)
+		nbytes = 8192;
+
+	if (!(buffer = (char *)__get_free_page(GFP_KERNEL)))
+		return 0;
+
+	offset = sect->sh_offset;
+	while (nbytes > 0) {
+		char *p;
+		unsigned long count, str_start;
+
+#ifdef STACK_TOP
+		status = read_exec(exe_bprm->inode, offset, buffer,
+				nbytes > PAGE_SIZE ? PAGE_SIZE : nbytes, 1);
+#else
+		old_fs = get_fs();
+		set_fs (get_ds());
+		status = read_exec(exe_bprm->inode, offset, buffer,
+				nbytes > PAGE_SIZE ? PAGE_SIZE : nbytes);
+		set_fs(old_fs);
+#endif
+#if 0
+		if (ibcs_trace & TRACE_ELF_LD)
+			printk(KERN_DEBUG
+				"ELF: read %d bytes, offset %d, got %d\n",
+				nbytes, offset, status);
+#endif
+
+		if (status <= 0) {
+			free_page((unsigned long)buffer);
+			return 0;
+		}
+
+		p = buffer;
+		str_start = 0;
+		for (count=0; count<status; count++) {
+			struct elf_clue *clue;
+			char c = *(buffer + PAGE_SIZE - 1);
+			*(buffer + PAGE_SIZE - 1) = '\0';
+#if 0
+			if (ibcs_trace & TRACE_ELF_LD)
+				printk(KERN_DEBUG "ELF: testing %s\n", p);
+#endif
+			for (clue=elf_clues; clue->len; clue++) {
+				if ((clue->len < 0 && strstr(p, clue->text))
+				|| (clue->len > 0 && !strncmp(p, clue->text, clue->len))) {
+#ifdef ELF_TRACE
+					if (ibcs_trace & TRACE_ELF_LD) {
+						printk(KERN_DEBUG
+							"ELF: testing %s\n",
+							p);
+						printk(KERN_DEBUG
+							"ELF:    with %s\n",
+							clue->text);
+						printk(KERN_DEBUG
+							"ELF:  giving 0x%08lx\n",
+							clue->personality);
+					}
+#endif
+					current->personality = clue->personality;
+					free_page((unsigned long)buffer);
+					return (1);
+				}
+			}
+			*(buffer + PAGE_SIZE - 1) = c;
+
+			while (*p && count < status) p++,count++;
+			if (count < status) {
+				p++;
+				count++;
+				str_start = count;
+			}
+		}
+
+		/* If we didn't find an end ofstring at all this page
+		 * probably isn't useful string data.
+		 */
+		if (str_start == 0)
+			str_start = status;
+
+		nbytes -= str_start;
+		offset += str_start;
+	}
+
+	free_page((unsigned long)buffer);
+	return (0);
+}
+#endif /* ELF_PARSE_COMMENTS */
+
+
 /* This is really simpleminded and specialized - we are loading an
    a.out library that is given an ELF header. */
 
@@ -762,14 +1015,19 @@ static int load_elf_library(int fd){
 	inode = file->f_inode;
 	elf_bss = 0;
 	
-	set_fs(KERNEL_DS);
-	if (file->f_op->read(inode, file, (char *) &elf_ex, sizeof(elf_ex)) != sizeof(elf_ex)) {
-		SYS(close)(fd);
+	if (SYS(lseek)(fd, 0, 0) != 0) {
 		MOD_DEC_USE_COUNT;
-		return -EACCES;
+		return -ENOEXEC;
 	}
+
+	set_fs(KERNEL_DS);
+	error = file->f_op->read(inode, file, (char *) &elf_ex, sizeof(elf_ex));
 	set_fs(USER_DS);
-	
+	if (error != sizeof(elf_ex)) {
+		MOD_DEC_USE_COUNT;
+		return -ENOEXEC;
+	}
+
 	if (elf_ex.e_ident[0] != 0x7f ||
 	    strncmp(&elf_ex.e_ident[1], "ELF",3) != 0) {
 		MOD_DEC_USE_COUNT;
