@@ -6,8 +6,8 @@
  *  Written by Drew Sullivan.
  *  Rewritten by Mike Jagdis.
  *
- * $Id: ioctl.c,v 1.38 1995/07/17 09:22:39 mike Exp $
- * $Source: /u3/CVS/ibcs/iBCSemul/ioctl.c,v $
+ * $Id: ioctl.c,v 1.47 1996/06/06 16:23:02 mike Exp $
+ * $Source: /usr/CVS/ibcs/iBCSemul/ioctl.c,v $
  */
 #include <linux/config.h>
 
@@ -25,6 +25,7 @@
 #include <linux/sched.h>
 #include <linux/kernel.h>
 #include <linux/termios.h>
+#include <linux/mtio.h>
 #include <linux/time.h>
 #include <linux/sockios.h>
 #include <linux/mm.h>
@@ -39,6 +40,8 @@
 
 
 static int bsd_ioctl_file(int fd, unsigned int func, void *arg);
+static int sco_ioctl_tape(int fd, unsigned int func, unsigned long arg);
+static int ibcs_ioctl_tape(int fd, unsigned int func, unsigned long arg);
 static int ibcs_ioctl_termios(int fd, unsigned int func, void *arg);
 static int ibcs_ioctl_termio(int fd, unsigned int func, void *arg);
 static int ibcs_ioctl_console(int fd, unsigned int func, void *arg);
@@ -88,6 +91,15 @@ do_ioctl(struct pt_regs *regs, int fd, unsigned long ioctl_num, void *arg)
 		case 'A': /* ??? SCO console keyboard stuff? */
 			return -EINVAL;
 
+#ifdef EMU_SCO
+		case 'm':
+			return sco_ioctl_tape(fd, ioctl_num, (unsigned long)arg);
+#endif
+#ifdef EMU_SVR4
+		case 't':
+			return ibcs_ioctl_tape(fd, ioctl_num, (unsigned long)arg);
+#endif
+
 		case 'f':
 			return bsd_ioctl_file(fd, ioctl_num, arg);
 
@@ -192,7 +204,7 @@ do_ioctl(struct pt_regs *regs, int fd, unsigned long ioctl_num, void *arg)
 	/* If nothing has handled it yet someone may have to do some
 	 * more work...
 	 */
-	printk(KERN_ERR "iBCS: ioctl(%d, %x[%s], 0x%lx) unsupported\n",
+	printk(KERN_ERR "iBCS: ioctl(%d, %lx[%s], 0x%lx) unsupported\n",
 		fd, ioctl_num, fix(class), (unsigned long)arg);
 
 	return -EINVAL;
@@ -529,24 +541,27 @@ static int svr4_to_linux_termios(int fd, int op, struct svr4_termios *it)
 	t.c_lflag &= ~0001000;
 	t.c_lflag |= (lflag & 0001000); /* Restore ECHOCTL */
 
-	memcpy_fromfs(svr4_cc, &it->c_cc, SCO_NCCS);
+	memcpy_fromfs(svr4_cc, &it->c_cc, SVR4_NCCS);
 	t.c_cc[0] = svr4_cc[0];
 	t.c_cc[1] = svr4_cc[1];
 	t.c_cc[2] = svr4_cc[2];
 	t.c_cc[3] = svr4_cc[3];
 	t.c_cc[7] = svr4_cc[7];
-	t.c_cc[8] = svr4_cc[11];
-	t.c_cc[9] = svr4_cc[12];
+	t.c_cc[8] = svr4_cc[8];
+	t.c_cc[9] = svr4_cc[9];
 	t.c_cc[10] = svr4_cc[10];
+	t.c_cc[12] = svr4_cc[12];
+	t.c_cc[13] = svr4_cc[13];
+	t.c_cc[14] = svr4_cc[14];
+	t.c_cc[15] = svr4_cc[15];
 	t.c_cc[16] = svr4_cc[16];
 	if (t.c_lflag & ICANON) {
 		t.c_cc[4] = svr4_cc[4];
 		t.c_cc[11] = svr4_cc[5];
 	} else {
-		t.c_cc[4] = svr4_cc[8];
 		t.c_cc[5] = svr4_cc[5];
 		t.c_cc[6] = svr4_cc[4];
-		t.c_cc[11] = svr4_cc[9];
+		t.c_cc[11] = svr4_cc[6];
 	}
 
 	set_fs(get_ds());
@@ -595,11 +610,14 @@ static int linux_to_svr4_termios(int fd, int op, struct svr4_termios *it)
 	svr4_cc[3] = t.c_cc[3];
 	svr4_cc[6] = t.c_cc[16];
 	svr4_cc[7] = t.c_cc[7];
-	svr4_cc[8] = t.c_cc[4];
-	svr4_cc[9] = t.c_cc[11];
+	svr4_cc[8] = t.c_cc[8];
+	svr4_cc[9] = t.c_cc[9];
 	svr4_cc[10] = t.c_cc[10];
-	svr4_cc[11] = t.c_cc[8];
-	svr4_cc[12] = t.c_cc[9];
+	svr4_cc[11] = t.c_cc[10];
+	svr4_cc[12] = t.c_cc[12];
+	svr4_cc[13] = t.c_cc[13];
+	svr4_cc[14] = t.c_cc[14];
+	svr4_cc[15] = t.c_cc[15];
 	if (t.c_lflag & ICANON) {
 		svr4_cc[4] = t.c_cc[4];
 		svr4_cc[5] = t.c_cc[11];
@@ -768,7 +786,7 @@ static int ibcs_ioctl_termiox(int fd, unsigned int func, void *arg) {
 	}
 
 	memcpy_fromfs(&tx, arg, sizeof(struct termiox));
-	if ((tx.x_hflag != 0 && tx.x_hflag != RTSXOFF|CTSXON)
+	if ((tx.x_hflag != 0 && tx.x_hflag != (RTSXOFF|CTSXON))
 	|| tx.x_cflag || tx.x_rflag[0] || tx.x_rflag[1]
 	|| tx.x_rflag[2] || tx.x_rflag[3] || tx.x_rflag[4]
 	|| tx.x_sflag)
@@ -809,6 +827,15 @@ static int ibcs_ioctl_stream(struct pt_regs *regs,
 					arg, sizeof(unsigned long));
 			if (error)
 				return error;
+#ifdef EMU_XTI
+			if (!current->FD[fd])
+				return -EBADF;
+			if (current->FD[fd]->f_inode->i_sock
+			&& Priv(fd) && Priv(fd)->pfirst) {
+				put_fs_long(Priv(fd)->pfirst->length, arg);
+				return 1; /* at least 1... (FIXME) */
+			}
+#endif
 			error = SYS(ioctl)(fd, FIONREAD, arg);
 			if (error)
 				return error;
@@ -896,14 +923,58 @@ static int ibcs_ioctl_stream(struct pt_regs *regs,
 			return 0;
 		}
 
+		case 011: /* I_SETSIG */
+			return SYS(ioctl)(fd, FIOSETOWN, current->pid);
+		case 012: /* I_GETSIG */
+			return SYS(ioctl)(fd, FIOGETOWN, arg);
+
+		case 017: { /* I_PEEK */
+#ifdef EMU_XTI
+			struct strpeek buf;
+
+			error = verify_area(VERIFY_WRITE,
+					arg, sizeof(struct strpeek));
+			if (error)
+				return error;
+			memcpy_fromfs(&buf, arg, sizeof(buf));
+
+			if (!current->FD[fd])
+				return -EBADF;
+
+			if (current->FD[fd]->f_inode->i_sock
+			&& Priv(fd) && Priv(fd)->pfirst
+			&& (!buf.flags || buf.flags == Priv(fd)->pfirst->pri)) {
+				int l = buf.ctl.maxlen <= Priv(fd)->pfirst->length
+					? buf.ctl.maxlen
+					: Priv(fd)->pfirst->length;
+				error = verify_area(VERIFY_WRITE,
+						buf.ctl.buf, l);
+				if (error)
+					return error;
+				memcpy_tofs(buf.ctl.buf,
+					((char *)&Priv(fd)->pfirst->type)
+					+ Priv(fd)->offset, l);
+				put_fs_long(l,
+					&((struct strpeek *)arg)->ctl.len);
+				if (buf.dat.maxlen >= 0)
+					put_fs_long(0,
+						&((struct strpeek *)arg)->dat.len);
+				put_fs_long(Priv(fd)->pfirst->pri,
+					&((struct strpeek *)arg)->flags);
+				return 1;
+			}
+			/* FIXME: I think we should also be able to peek
+			 * at data as well?
+			 */
+#endif
+			return 0; /* Nothing to peek at. */
+		}
+
 		case 004: /* I_LOOK */
 		case 006: /* I_SRDOPT */
 		case 007: /* I_GRDOPT */
-		case 011: /* I_SETSIG */
-		case 012: /* I_GETSIG */
 		case 014: /* I_LINK */
 		case 015: /* I_UNLINK */
-		case 017: /* I_PEEK */
 		case 020: /* I_FDINSERT */
 		case 021: /* I_SENDFD */
 		case 022: /* I_RECVFD */
@@ -933,6 +1004,201 @@ static int ibcs_ioctl_stream(struct pt_regs *regs,
 }
 
 
+#ifdef EMU_SCO
+static int sco_ioctl_tape(int fd, unsigned int func, unsigned long arg)
+{
+	int old_fs;
+	int error;
+	struct mtop mtop;
+
+	mtop.mt_count = 1;
+
+	switch (func & 0xff) {
+		case 1:  /* MT_RESET */
+			mtop.mt_op = MTRESET;
+			break;
+
+		case 2:  /* MT_RETEN */
+			mtop.mt_op = MTRETEN;
+			break;
+
+		case 3:  /* MT_REWIND */
+			mtop.mt_op = MTREW;
+			break;
+
+		case 4:  /* MT_ERASE */
+		case 23:  /* HP_ERASE */
+			mtop.mt_op = MTERASE;
+			break;
+
+		case 6:  /* MT_RFM */
+			mtop.mt_op = MTFSF;
+			break;
+
+		case 7:  /* MT_WFM */
+			mtop.mt_op = MTWEOF;
+			break;
+
+		case 8:  /* MT_LOAD */
+			mtop.mt_op = MTLOAD;
+			break;
+
+		case 9:  /* MT_UNLOAD */
+			mtop.mt_op = MTOFFL;
+			break;
+
+		case 19:  /* MT_RSM */
+			mtop.mt_op = MTFSS;
+			break;
+
+		case 20:  /* MT_WSM */
+			mtop.mt_op = MTWSM;
+			break;
+
+		case 21:  /* MT_EOD */
+			mtop.mt_op = MTEOM;
+			break;
+
+		case 25:  /* MT_LOCK */
+			mtop.mt_op = MTLOCK;
+			break;
+
+		case 26:  /* MT_UNLOCK */
+			mtop.mt_op = MTUNLOCK;
+			break;
+
+
+#if 0
+/*  The following function codes are just copied from the SCO
+    include file.
+*/
+		case 0:  /* MT_STATUS */
+		case 5:  /* MT_AMOUNT */
+		case 10:  /* MT_DSTATUS */
+		case 11:  /* MT_FORMAT */
+		case 12:  /* MT_GETHDR */
+		case 13:  /* MT_PUTHDR */
+		case 14:  /* MT_GETNEWBB */
+		case 15:  /* MT_PUTNEWBB */
+		case 16:  /* MT_GETVTBL */
+		case 17:  /* MT_PUTVTBL */
+		case 18:  /* MT_SERVO */
+		case 22:  /* MT_FORMPART */
+		case 24:  /* MT_SETBLK */
+		case 38:  /* MT_SETANSI */
+		case 64:  /* MT_REPORT */
+#endif
+		default:
+			printk (KERN_ERR "iBCS: SCO tape ioctl func=%d arg=%x unsupported\n",
+				func & 0xff, (int) arg);
+			return -EINVAL;
+	}
+
+	old_fs = get_fs ();
+	set_fs (get_ds ());
+	error = SYS (ioctl) (fd, MTIOCTOP, &mtop);
+	set_fs (old_fs);
+	return error;
+}
+#endif
+
+#ifdef EMU_SVR4
+static int ibcs_ioctl_tape(int fd, unsigned int func, unsigned long arg)
+{
+	int old_fs;
+	int error;
+	struct mtop mtop;
+
+	mtop.mt_count = 1;
+
+	switch (func & 0xff) {
+		case 1:		/* MT_RETEN */
+			mtop.mt_op = MTRETEN;
+			break;
+
+		case 2:		/* MT_REWIND */
+			mtop.mt_op = MTREW;
+			break;
+
+		case 3:		/* MT_ERASE */
+			mtop.mt_op = MTERASE;
+			break;
+
+		case 4:		/* MT_WFM */
+			mtop.mt_op = MTWEOF;
+			break;
+
+		case 5:		/* MT_RESET */
+			mtop.mt_op = MTRESET;
+			break;
+
+		case 7:		/* T_SFF */
+			mtop.mt_op = MTFSF;
+			break;
+
+		case 8:		/* T_SBF */
+			mtop.mt_op = MTBSF;
+			break;
+
+		case 9:		/* T_LOAD */
+			mtop.mt_op = MTLOAD;
+			break;
+
+		case 10:  /* MT_UNLOAD */
+			mtop.mt_op = MTOFFL;
+			break;
+
+		case 16:	/* T_PREVMV */
+			mtop.mt_op = MTLOCK;
+			break;
+
+		case 17:	/* T_ALLOMV */
+			mtop.mt_op = MTUNLOCK;
+			break;
+
+		case 20:	/* T_EOD */
+			mtop.mt_count = arg;
+			mtop.mt_op = MTEOM;
+			break;
+
+		case 21:	/* T_SSFB */
+			mtop.mt_count = arg;
+			mtop.mt_op = MTBSFM;
+			break;
+
+		case 22:	/* T_SSFF */
+			mtop.mt_count = arg;
+			mtop.mt_op = MTFSFM;
+			break;
+
+		case 24:	/* T_STD */
+			mtop.mt_count = arg;
+			mtop.mt_op = MTSETDENSITY;
+			break;
+
+#if 0
+		case 6:		/* T_STATUS */
+		case 14:	/* T_RDBLKLEN */
+		case 15:	/* T_WRBLKLEN */
+		case 18:	/* T_SBB */
+		case 19:	/* T_SFB */
+		case 23:	/* T_STS */
+#endif
+		default:
+			printk (KERN_ERR "iBCS: SYSV tape ioctl func=%d arg=%x unsupported\n",
+				func & 0xff, (int) arg);
+			return -EINVAL;
+	}
+
+	old_fs = get_fs ();
+	set_fs (get_ds ());
+	error = SYS (ioctl) (fd, MTIOCTOP, &mtop);
+	set_fs (old_fs);
+	return error;
+}
+#endif
+
+
 int
 ibcs_ioctl(struct pt_regs *regs)
 {
