@@ -26,5 +26,13 @@ int ibcs_mmap(unsigned int vaddr, unsigned int vsize, int prot, int flags,
 		if (fd >= NR_OPEN || !(file = current->FD[fd]))
 			return -EBADF;
 	}
-	return do_mmap(file, vaddr, vsize, prot, flags, file_offset);
+
+	if (current->personality == PER_SVR4
+	&& !(flags & 0x80000000) && vaddr) {
+		unsigned int ret;
+		ret = do_mmap(file, vaddr, vsize, prot, flags | MAP_FIXED, file_offset);
+		return (ret == vaddr ? 0 : ret);
+	}
+
+	return do_mmap(file, vaddr, vsize, prot, flags & 0x7fffffff, file_offset);
 }
