@@ -11,8 +11,8 @@
  *  Please consider this closely if you plan on changing this mode.
  *  -- Al Longyear
  *
- * $Id: signal.c,v 1.20 1995/06/22 12:11:29 mike Exp $
- * $Source: /u3/CVS/ibcs/iBCSemul/signal.c,v $
+ * $Id: signal.c,v 1.23 1995/11/29 13:00:39 mike Exp $
+ * $Source: /usr/CVS/ibcs/iBCSemul/signal.c,v $
  */
 
 #include <linux/config.h>
@@ -160,6 +160,7 @@ void ibcs_sig_handler (struct pt_regs * regs, int sig,
 	int	      old_fs;
 
 	sigemptyset (&act.sa_mask);
+	act.sa_restorer = NULL;
 	act.sa_handler = handler;
 	act.sa_flags   = 0;
 
@@ -314,6 +315,7 @@ void ibcs_sigignore (struct pt_regs * regs)
 
 	sigemptyset (&act.sa_mask);
 
+	act.sa_restorer = NULL;
 	act.sa_handler = SIG_IGN;
 	act.sa_flags   = 0;
 
@@ -339,52 +341,24 @@ void ibcs_sigignore (struct pt_regs * regs)
 
 void ibcs_sigpause (struct pt_regs * regs)
 {
-	sigset_t   newset, oldset, zeroset, pendingset;
+	sigset_t   newset;
 	int	sig, answer;
-	int	old_fs;
 
-	if (!ibcs_signo (regs, &sig))
+	if (!ibcs_signo(regs, &sig))
 		return;
 
-	sigemptyset (&zeroset);
-	sigemptyset (&newset);
-	sigaddset   (&newset, sig);
-
-	TO_KERNEL (old_fs);
-	answer    = SYS(sigprocmask) (SIG_BLOCK, &newset, &oldset);
-	FROM_KERNEL (old_fs);
-
-	if (answer >= 0) {
-		TO_KERNEL (old_fs);
-		answer = SYS(sigpending) (&pendingset);
-		FROM_KERNEL (old_fs);
-
-		while (answer >= 0) {
-			if (sigismember (&pendingset, sig) != 0)
-				break;
-
-			answer = SYS(sigsuspend)(regs->ebx, regs->ecx,
-					regs->edx, regs->esi, regs->edi,
-					regs->ebp, regs->eax,
-					regs->ds, regs->es,
-					regs->fs, regs->gs, regs->orig_eax,
-					regs->eip, regs->cs, regs->eflags,
-					regs->esp, regs->ss,
-					zeroset);
-
-			if (answer >= 0) {
-				TO_KERNEL (old_fs);
-				answer = SYS(sigpending) (&pendingset);
-				FROM_KERNEL (old_fs);
-			}
-		}
-		TO_KERNEL (old_fs);
-		SYS(sigprocmask) (SIG_SETMASK, &oldset, NULL);
-		FROM_KERNEL (old_fs);
-	}
+	sigfillset(&newset);
+	sigdelset(&newset, sig);
+	answer = SYS(sigsuspend)(0, current->blocked,
+			newset, regs->esi, regs->edi,
+			regs->ebp, regs->eax,
+			regs->ds, regs->es,
+			regs->fs, regs->gs, regs->orig_eax,
+			regs->eip, regs->cs, regs->eflags,
+			regs->esp, regs->ss);
 
 	if (answer < 0) {
-		regs->eax     = iABI_errors (-answer);
+		regs->eax = iABI_errors(-answer);
 		regs->eflags |= 1;
 	}
 }
@@ -398,55 +372,49 @@ void ibcs_sigpause (struct pt_regs * regs)
 
 int ibcs_sigfunc (struct pt_regs * regs)
 {
-#ifdef IBCS_TRACE
-        char *func;
-#endif
 	int sig_type = (int) HIDDEN_PARAM;
 
+#ifdef IBCS_TRACE
+	if ((ibcs_trace & (TRACE_SIGNAL | TRACE_SIGNAL_F))
+	|| ibcs_func_p->trace) {
+		printk(KERN_DEBUG "iBCS2 sig%s(%ld, 0x%08lx, 0x%08lx)\n",
+			sig_type == 0 ? "nal"
+			: (sig_type == 0x100 ? "set"
+			: (sig_type == 0x200 ? "hold"
+			: (sig_type == 0x400 ? "relse"
+			: (sig_type == 0x800 ? "ignore"
+			: (sig_type == 0x1000 ? "pause"
+			: "???" ))))),
+			SIGNAL_NUMBER & 0xff, SECOND_PARAM, THIRD_PARAM);
+	}
+#endif
+
 	regs->eflags &= ~1;
 	regs->eax     = 0;
 
 	switch (sig_type) {
 	case 0x0000:
 		ibcs_signal (regs);
-#ifdef IBCS_TRACE
-		func = "nal";
-#endif
 		break;
 
 	case 0x0100:
 		ibcs_sigset (regs);
-#ifdef IBCS_TRACE
-		func = "set";
-#endif
 		break;
 
 	case 0x0200:
 		ibcs_sighold (regs);
-#ifdef IBCS_TRACE
-		func = "hold";
-#endif
 		break;
 		
 	case 0x0400:
 		ibcs_sigrelse (regs);
-#ifdef IBCS_TRACE
-		func = "relse";
-#endif
 		break;
 
 	case 0x0800:
 		ibcs_sigignore (regs);
-#ifdef IBCS_TRACE
-		func = "nore";
-#endif
 		break;
 
 	case 0x1000:
 		ibcs_sigpause (regs);
-#ifdef IBCS_TRACE
-		func = "pause";
-#endif
 		break;
 
 	default:
@@ -468,9 +436,7 @@ int ibcs_sigfunc (struct pt_regs * regs)
 #ifdef IBCS_TRACE
 	if ((ibcs_trace & (TRACE_SIGNAL | TRACE_SIGNAL_F))
 	|| ibcs_func_p->trace) {
-		printk (KERN_DEBUG "iBCS2 sig%s(%ld, 0x%08lx, 0x%08lx) = %ld\n",
-			func, SIGNAL_NUMBER & 0xff, SECOND_PARAM, THIRD_PARAM,
-			regs->eax);
+		printk(KERN_DEBUG "iBCS2 returns %ld\n", regs->eax);
 	}
 #endif
 	return 0;
@@ -507,7 +473,7 @@ asmlinkage int abi_sigaction(int abi_signum, const struct abi_sigaction * action
 	signum = ibcs_mapsig(abi_signum);
 	if (signum<1 || signum>32 || signum==SIGKILL || signum==SIGSTOP)
 		return -EINVAL;
-	p = signum - 1 + current->sigaction;
+	p = signum - 1 + current->SIGACTION;
 	if (action) {
 		memcpy_fromfs(&new_sa, action, sizeof(struct abi_sigaction));
 		if (new_sa.sa_flags & ABI_SA_NODEFER)
@@ -546,6 +512,7 @@ asmlinkage int abi_sigaction(int abi_signum, const struct abi_sigaction * action
 	if (action) {
 	        /* The internal format of the sigaction structure is
 		   different, so we cannot simply copy the structure. */
+		p->sa_restorer = NULL;
 	  	p->sa_handler = new_sa.sa_handler;
 #ifdef INIT_MM
 		  p->sa_mask = map_bitvec(new_sa.sa_mask,
@@ -595,7 +562,7 @@ asmlinkage int sco_sigaction(int sco_signum, const struct sco_sigaction * action
 	signum = ibcs_mapsig(sco_signum);
 	if (signum<1 || signum>32 || signum==SIGKILL || signum==SIGSTOP)
 		return -EINVAL;
-	p = signum - 1 + current->sigaction;
+	p = signum - 1 + current->SIGACTION;
 	if (action) {
 		memcpy_fromfs(&new_sa, action, sizeof(struct sco_sigaction));
 		new_sa.sa_mask |= _S(sco_signum);
@@ -622,6 +589,7 @@ asmlinkage int sco_sigaction(int sco_signum, const struct sco_sigaction * action
 	if (action) {
 	        /* The internal format of the sigaction structure is
 		   different, so we cannot simply copy the structure. */
+		p->sa_restorer = NULL;
 	  	p->sa_handler = new_sa.sa_handler;
 #ifdef INIT_MM
 		  p->sa_mask = map_bitvec(new_sa.sa_mask,
@@ -728,12 +696,11 @@ int abi_sigsuspend(struct pt_regs * regs)
 		printk("iBCS: sigsuspend oldset, newset = %lx %lx\n",
 			oldset, newset);
 #endif
-	return SYS(sigsuspend)(regs->ebx, regs->ecx,
-			regs->edx, regs->esi, regs->edi,
+	return SYS(sigsuspend)(0, oldset,
+			newset, regs->esi, regs->edi,
 			regs->ebp, regs->eax,
 			regs->ds, regs->es,
 			regs->fs, regs->gs, regs->orig_eax,
 			regs->eip, regs->cs, regs->eflags,
-			regs->esp, regs->ss,
-			newset);
+			regs->esp, regs->ss);
 }
