@@ -1,15 +1,10 @@
 /*
  *  linux/ibcs/socksys.h
  *
- *  Copyright 1994, 1995  Mike Jagdis (jaggy@purplet.demon.co.uk)
+ *  Copyright 1994-1996  Mike Jagdis (jaggy@purplet.demon.co.uk)
  *
- * $Id: socksys.c,v 1.39 1995/07/03 12:51:52 mike Exp $
- * $Source: /u3/CVS/ibcs/iBCSemul/socksys.c,v $
- */
-
-/* This is designed to work with Net2Debugged. It requires patches to
- * Net2Debugged. It will almost certainly not work with Net2E without
- * a fair amount of work.
+ * $Id: socksys.c,v 1.45 1996/03/29 17:33:35 mike Exp $
+ * $Source: /usr/CVS/ibcs/iBCSemul/socksys.c,v $
  */
 
 #include <linux/config.h>
@@ -55,7 +50,27 @@
 
 static int socksys_open(struct inode *ino, struct file *filep);
 static void socksys_close(struct inode *ino, struct file *filep);
+static int socksys_select(struct inode *ino, struct file *filep,
+			int sel_type, select_table *wait);
 
+static int spx_write(struct inode *ino, struct file *filep,
+			const char *buf, int count);
+
+/* spx_fops defines the file operations that can be applied to the
+ * /dev/spx server devices.
+ */
+static struct file_operations spx_fops = {
+	NULL,			/* lseek */
+	NULL,			/* read */
+	spx_write,		/* write */
+	NULL,			/* readdir */
+	NULL,			/* select */
+	NULL,			/* ioctl - ibcs_ioctl calls ibcs_ioctl_socksys*/
+	NULL,			/* mmap */
+	NULL,			/* open */
+	socksys_close,		/* close */
+	NULL			/* fsync */
+};
 
 /* socksys_fops defines the file operations that can be applied to the
  * /dev/socksys device.
@@ -90,6 +105,33 @@ static struct file_operations socksys_socket_fops = {
 	NULL			/* fsync */
 };
 static void (*sock_close)(struct inode *inode, struct file *file);
+static int (*sock_select)(struct inode *ino, struct file *filep,
+			int sel_type, select_table *wait);
+
+
+static void
+inherit_socksys_funcs(struct file *filep, int fd)
+{
+	/* If our file operations don't appear to match
+	 * what the socket system is advertising it may
+	 * be because we haven't initialised ours at all
+	 * yet or it may be because the old socket system
+	 * module was unloaded and reloaded. This isn't
+	 * entirely safe because we may still have open
+	 * sockets which *should* use the old routines
+	 * until they close - tough, for now.
+	 */
+	if (socksys_socket_fops.read != current->FD[fd]->f_op->read) {
+		memcpy(&socksys_socket_fops,
+			current->FD[fd]->f_op,
+			sizeof(struct file_operations));
+		sock_close = socksys_socket_fops.release;
+		sock_select = socksys_socket_fops.select;
+		socksys_socket_fops.release = socksys_close;
+		socksys_socket_fops.select = socksys_select;
+	}
+	filep->f_op = &socksys_socket_fops;
+}
 
 
 static inline int
@@ -146,23 +188,7 @@ socksys_syscall(int fd, int *sp)
 			if ((fd = SYS(socketcall)(SYS_SOCKET, sp)) < 0)
 				return fd;
 
-			/* If our file operations don't appear to match
-			 * what the socket system is advertising it may
-			 * be because we haven't initialised ours at all
-			 * yet or it may be because the old socket system
-			 * module was unloaded and reloaded. This isn't
-			 * entirely safe because we may still have open
-			 * sockets which *should* use the old routines
-			 * until they close - tough, for now.
-			 */
-			if (socksys_socket_fops.read != current->FD[fd]->f_op->read) {
-				memcpy(&socksys_socket_fops,
-					current->FD[fd]->f_op,
-					sizeof(struct file_operations));
-				sock_close = socksys_socket_fops.release;
-				socksys_socket_fops.release = socksys_close;
-			}
-			current->FD[fd]->f_op = &socksys_socket_fops;
+			inherit_socksys_funcs(current->FD[fd], fd);
 			MOD_INC_USE_COUNT;
 			return fd;
 		}
@@ -173,23 +199,7 @@ socksys_syscall(int fd, int *sp)
 			if ((fd = SYS(socketcall)(SYS_ACCEPT, sp)) < 0)
 				return fd;
 
-			/* If our file operations don't appear to match
-			 * what the socket system is advertising it may
-			 * be because we haven't initialised ours at all
-			 * yet or it may be because the old socket system
-			 * module was unloaded and reloaded. This isn't
-			 * entirely safe because we may still have open
-			 * sockets which *should* use the old routines
-			 * until they close - tough, for now.
-			 */
-			if (socksys_socket_fops.read != current->FD[fd]->f_op->read) {
-				memcpy(&socksys_socket_fops,
-					current->FD[fd]->f_op,
-					sizeof(struct file_operations));
-				sock_close = socksys_socket_fops.release;
-				socksys_socket_fops.release = socksys_close;
-			}
-			current->FD[fd]->f_op = &socksys_socket_fops;
+			inherit_socksys_funcs(current->FD[fd], fd);
 			MOD_INC_USE_COUNT;
 			return fd;
 		}
@@ -636,14 +646,112 @@ ibcs_ioctl_socksys(int fd, unsigned int cmd, void *arg)
 }
 
 
+static int
+spx_write(struct inode *ino, struct file *filep, const char *buf, int count)
+{
+	int newfd, err, args[3];
+	struct sockaddr_un addr = {
+		AF_UNIX, "/tmp/.X11-unix/X0"
+	};
+	int old_fs = get_fs();
+
+	newfd = get_fs_byte(buf);
+
+#ifdef IBCS_TRACE
+	if (ibcs_trace & TRACE_SOCKSYS)
+		printk(KERN_DEBUG "SPX: [%d] %lx choose service %d\n",
+			current->pid, (unsigned long)filep, newfd);
+#endif
+
+	/* Rather than use an explicit path to the X :0 server
+	 * socket we should use the given number to look up a path
+	 * name to use (we can't rely on servers registering their
+	 * sockets either - for one thing we don't emulate that yet
+	 * and for another thing different OS binaries do things in
+	 * different ways but all must interoperate).
+	 * I suggest putting the mapping in, say, /dev/spx.map/%d
+	 * where each file is a symlink containing the path of the
+	 * socket to use. Then we can just do a readlink() here to
+	 * get the pathname.
+	 *   Hey, this is what we do here now!
+	 */
+	addr.sun_family = AF_UNIX;
+	sprintf(addr.sun_path, "/dev/spx.map/%d", newfd);
+	set_fs(get_ds());
+	err = SYS(readlink)(addr.sun_path, addr.sun_path,
+				sizeof(addr.sun_path)-1);
+	set_fs(old_fs);
+	if (err == -ENOENT) {
+#ifdef IBCS_TRACE
+		if (ibcs_trace & TRACE_SOCKSYS)
+			printk(KERN_DEBUG
+				"SPX: [%d] %lx no symlink \"%s\", try X :0\n",
+				current->pid, (unsigned long)filep,
+				addr.sun_path);
+#endif
+		strcpy(addr.sun_path, "/tmp/.X11-unix/X0");
+	} else {
+		if (err < 0)
+			return err;
+		addr.sun_path[err] = '\0';
+	}
+
+#ifdef IBCS_TRACE
+	if (ibcs_trace & TRACE_SOCKSYS)
+		printk(KERN_DEBUG "SPX: [%d] %lx get a Unix domain socket\n",
+			current->pid, (unsigned long)filep);
+#endif
+	args[0] = AF_UNIX;
+	args[1] = SOCK_STREAM;
+	args[2] = 0;
+	set_fs(get_ds());
+	newfd = SYS(socketcall)(SYS_SOCKET, args);
+	set_fs(old_fs);
+	if (newfd < 0)
+		return newfd;
+
+#ifdef IBCS_TRACE
+	if (ibcs_trace & TRACE_SOCKSYS)
+		printk(KERN_DEBUG "SPX: [%d] %lx connect to \"%s\"\n",
+			current->pid, (unsigned long)filep, addr.sun_path);
+#endif
+	args[0] = newfd;
+	args[1] = (int)&addr;
+	args[2] = sizeof(struct sockaddr_un);
+	set_fs(get_ds());
+	err = SYS(socketcall)(SYS_CONNECT, args);
+	set_fs(old_fs);
+	if (err) {
+		SYS(close)(newfd);
+		return err;
+	}
+
+#ifdef IBCS_TRACE
+	if (ibcs_trace & TRACE_SOCKSYS)
+		printk(KERN_DEBUG "SPX: [%d] filep=0x%08lx, swap inodes old=0x%08lx, new=0x%08lx\n",
+			current->pid, (unsigned long)filep, (unsigned long)ino,
+			(unsigned long)current->FD[newfd]->f_inode);
+#endif
+	inherit_socksys_funcs(filep, newfd);
+	current->FD[newfd]->f_inode->i_rdev = ino->i_rdev;
+	iput(ino);
+	filep->f_inode = current->FD[newfd]->f_inode;
+	if (--current->FD[newfd]->f_count == 0)
+		current->FD[newfd] = NULL;
+
+	return 1;
+}
+
+
 static int
 socksys_open(struct inode *ino, struct file *filep)
 {
 	MOD_INC_USE_COUNT;
 #ifdef IBCS_TRACE
 	if (ibcs_trace & TRACE_SOCKSYS)
-		printk(KERN_DEBUG "socksys: [%d] %lx opening\n",
-			current->pid, (unsigned long)filep);
+		printk(KERN_DEBUG
+			"socksys: [%d] filep=0x%08lx, inode=0x%08lx opening\n",
+			current->pid, (unsigned long)filep, (unsigned long)ino);
 #endif
 	/* Minor = 0 is the socksys device itself. No special handling
 	 *           will be needed as it is controlled by the application
@@ -660,7 +768,13 @@ socksys_open(struct inode *ino, struct file *filep)
 	 *           side (/dev/X0R - which must exist but can be a link
 	 *           to /dev/null) nor do we need to actually implement
 	 *           getmsg/putmsg.
-	 *
+	 */
+	if (MINOR(ino->i_rdev) == 1) {
+		filep->f_op = &spx_fops;
+		return 0;
+	}
+
+	/*
 	 * Otherwise the high 4 bits specify the address/protocol
 	 * family (AF_INET, AF_UNIX etc.) and the low 4 bits determine
 	 * the protocol (IPPROTO_IP, IPPROTO_UDP, IPPROTO_TCP etc.)
@@ -673,31 +787,19 @@ socksys_open(struct inode *ino, struct file *filep)
 	struct T_private *priv;
 #endif
 	int fd, err, args[3];
-	struct sockaddr_un Xaddr = {
-		AF_UNIX, "/tmp/.X11-unix/X0"
-	};
 	int old_fs = get_fs();
 
-#ifdef IBCS_TRACE
-	if (ibcs_trace & TRACE_SOCKSYS)
-		printk(KERN_DEBUG "XTI: [%d] %lx opening\n",
-			current->pid, (unsigned long)filep);
-#endif
-
 	/* Grab a socket. */
 #ifdef IBCS_TRACE
 	if (ibcs_trace & TRACE_SOCKSYS)
-		if (MINOR(ino->i_rdev) == 1)
-			printk(KERN_DEBUG "XTI: [%d] %lx get a Unix domain socket\n",
-				current->pid, (unsigned long)filep);
-		else
-			printk(KERN_DEBUG "TLI: [%d] %lx get socket for transport end point\n",
-				current->pid, (unsigned long)filep);
+		printk(KERN_DEBUG
+			"XTI: [%d] %lx get socket for transport end point"
+			" (dev = 0x%04x)\n",
+			current->pid, (unsigned long)filep,
+			ino->i_rdev);
 #endif
-	if (MINOR(ino->i_rdev) == 1) {
-	} else switch ((args[0] = ((MINOR(ino->i_rdev) >> 4) & 0x0f))) {
-		case 0:
-			/* Special - minor 1 is fake spx for X. */
+	switch ((args[0] = ((MINOR(ino->i_rdev) >> 4) & 0x0f))) {
+		case AF_UNIX:
 			args[0] = AF_UNIX;
 			args[1] = SOCK_STREAM;
 			args[2] = 0;
@@ -727,6 +829,13 @@ socksys_open(struct inode *ino, struct file *filep)
 			args[2] = 0;
 			break;
 	}
+#ifdef IBCS_TRACE
+	if (ibcs_trace & TRACE_SOCKSYS)
+		printk(KERN_DEBUG
+			"XTI: [%d] %lx socket %d %d %d\n",
+			current->pid, (unsigned long)filep,
+			args[0], args[1], args[2]);
+#endif
 	set_fs(get_ds());
 	fd = SYS(socketcall)(SYS_SOCKET, args);
 	set_fs(old_fs);
@@ -735,26 +844,6 @@ socksys_open(struct inode *ino, struct file *filep)
 		return fd;
 	}
 
-	if (MINOR(ino->i_rdev) == 1) {
-		/* Connect the socket to X. */
-#ifdef IBCS_TRACE
-		if (ibcs_trace & TRACE_SOCKSYS)
-			printk(KERN_DEBUG "XTI: [%d] %lx connect to /tmp/X11-unix/X0\n",
-				current->pid, (unsigned long)filep);
-#endif
-		set_fs (get_ds());
-		args[0] = fd;
-		args[1] = (int)&Xaddr;
-		args[2] = sizeof(struct sockaddr_un);
-		err = SYS(socketcall)(SYS_CONNECT, args);
-		set_fs(old_fs);
-		if (err) {
-			SYS(close)(fd);
-			MOD_DEC_USE_COUNT;
-			return err;
-		}
-	}
-
 #ifdef EMU_XTI
 	priv = (struct T_private *)kmalloc(sizeof(struct T_private), GFP_KERNEL);
 	if (!priv) {
@@ -762,28 +851,11 @@ socksys_open(struct inode *ino, struct file *filep)
 		MOD_DEC_USE_COUNT;
 		return err;
 	}
-	priv->state = T_UNBND;
+	priv->state = TS_UNBND;
 	priv->offset = 0;
 	priv->pfirst = priv->plast = NULL;
 #endif
 
-	/* If our file operations don't appear to match
-	 * what the socket system is advertising it may
-	 * be because we haven't initialised ours at all
-	 * yet or it may be because the old socket system
-	 * module was unloaded and reloaded. This isn't
-	 * entirely safe because we may still have open
-	 * sockets which *should* use the old routines
-	 * until they close - tough, for now.
-	 */
-	if (socksys_socket_fops.read != current->FD[fd]->f_op->read) {
-		memcpy(&socksys_socket_fops,
-			current->FD[fd]->f_op,
-			sizeof(struct file_operations));
-		sock_close = socksys_socket_fops.release;
-		socksys_socket_fops.release = socksys_close;
-	}
-
 	/* Release the inode we were given, replace with the socket
 	 * inode, redirect operations to our emulation handlers then
 	 * clear down the descriptor used for the socket.
@@ -793,9 +865,10 @@ socksys_open(struct inode *ino, struct file *filep)
 		printk(KERN_DEBUG "XTI: [%d] %lx swap inodes\n",
 			current->pid, (unsigned long)filep);
 #endif
+	inherit_socksys_funcs(filep, fd);
+	current->FD[fd]->f_inode->i_rdev = ino->i_rdev;
 	iput(ino);
 	filep->f_inode = current->FD[fd]->f_inode;
-	filep->f_op = &socksys_socket_fops;
 #ifdef EMU_XTI
 	filep->private_data = priv;
 #endif
@@ -807,6 +880,28 @@ socksys_open(struct inode *ino, struct file *filep)
 }
 
 
+static int
+socksys_select(struct inode *ino, struct file *filep,
+		int sel_type, select_table *wait)
+{
+#ifdef EMU_XTI
+	/* If we are checking for readable data and this is a
+	 * timod transport end point and there is a control
+	 * message queued we have readable data.
+	 */
+	if (sel_type == SEL_IN
+	&& ino && ino->i_sock && MINOR(ino->i_rdev) != 1
+	&& filep->private_data
+	&& ((struct T_private *)filep->private_data)->pfirst)
+		return 1;
+#endif
+
+	if (sock_select)
+		return (*sock_select)(ino, filep, sel_type, wait);
+	return 0;
+}
+
+
 static void
 socksys_close(struct inode *ino, struct file *filep)
 {
